from __future__ import annotations

import dataclasses
import sys
import typing

from . import serde as serde
from .generated.devices_types import *
from .mask import *
from .neuromorphic_drivers import list_devices as extension_list_devices
from .udev import *


@dataclasses.dataclass
class ListedDevice:
    name: Name
    speed: Speed
    serial: typing.Optional[str]
    error: typing.Optional[str]


def list_devices() -> list[ListedDevice]:
    return [
        ListedDevice(
            name=Name(name),
            speed=Speed(speed),
            serial=serial,
            error=error,
        )
        for (name, speed, serial, error) in extension_list_devices()
    ]


def print_device_list(color: bool = True):
    table = [["Name", "Speed", "Serial", "Error"]]
    for device in list_devices():
        table.append(
            [
                device.name.value,
                device.speed.value,
                "-" if device.serial is None else device.serial,
                "-" if device.error is None else device.error,
            ]
        )
    widths = [
        max(len(row[column]) for row in table) for column in range(0, len(table[0]))
    ]
    title = lambda string: f"\033[36;1m{string}\033[0m" if color else string
    sys.stdout.write(
        "".join(
            (
                "┌",
                "┬".join("─" * (width + 2) for width in widths),
                "┐\n",
                "".join(
                    (
                        "│",
                        "│".join(
                            title(f" {column:<{width}} ")
                            for width, column in zip(widths, table[0])
                        ),
                        "│",
                    )
                ),
                *(
                    (
                        "\n├",
                        "┼".join("─" * (width + 2) for width in widths),
                        "┤\n",
                        "\n".join(
                            "".join(
                                (
                                    "│",
                                    "│".join(
                                        f" {column:<{width}} "
                                        for width, column in zip(widths, row)
                                    ),
                                    "│",
                                )
                            )
                            for row in table[1:]
                        ),
                    )
                    if len(table) > 1
                    else ()
                ),
                "\n└",
                "┴".join("─" * (width + 2) for width in widths),
                "┘\n",
            )
        )
    )



import threading
import typing

import neuromorphic_drivers as nd
import numpy as np
import vispy.app
import vispy.gloo
import vispy.util.transforms

vispy.app.use_app(backend_name="glfw")

FRAME_DURATION: float = 1.0 / 60.0
ON_COLORMAP: list[str] = ["#F4C20D", "#191919"]
OFF_COLORMAP: list[str] = ["#1E88E5", "#191919"]

VERTEX_SHADER: str = """
uniform mat4 u_projection;
attribute vec2 a_position;
attribute vec2 a_texcoord;
varying vec2 v_texcoord;
void main (void) {
    v_texcoord = a_texcoord;
    gl_Position = u_projection * vec4(a_position, 0.0, 1.0);
}
"""


def color_to_vec4(color: str) -> str:
    return f"vec4({int(color[1:3], 16) / 255.0}, {int(color[3:5], 16) / 255.0}, {int(color[5:7], 16) / 255.0}, 1.0)"


FRAGMENT_SHADER: str = f"""
uniform sampler2D u_texture;
uniform float u_t;
uniform float u_tau;
const float on_color_table_scale = {len(ON_COLORMAP) - 1};
const vec4 on_color_table[2] = vec4[]({','.join(color_to_vec4(color) for color in ON_COLORMAP)});
const float off_color_table_scale = {len(OFF_COLORMAP) - 1};
const vec4 off_color_table[2] = vec4[]({','.join(color_to_vec4(color) for color in OFF_COLORMAP)});
varying vec2 v_texcoord;
void main() {{
    float t_and_on = texture2D(u_texture, v_texcoord).r;
    float lambda = 1.0f - exp(-float(u_t - abs(t_and_on)) / u_tau);
    float scaled_lambda = lambda * (t_and_on > 0.0 ? on_color_table_scale : off_color_table_scale);
    gl_FragColor = t_and_on == 0.0 ? on_color_table[int(on_color_table_scale)] : mix(
        (t_and_on > 0.0 ? on_color_table : off_color_table)[int(scaled_lambda)],
        (t_and_on > 0.0 ? on_color_table : off_color_table)[int(scaled_lambda) + 1],
        scaled_lambda - float(int(scaled_lambda))
    );
}}
"""


class Canvas(vispy.app.Canvas):
    def __init__(
        self, sensor_width: int, sensor_height: int, device: nd.GenericDeviceOptional
    ):
        self.program = None
        vispy.app.Canvas.__init__(
            self,
            keys="interactive",
            size=(sensor_width, sensor_height),
            vsync=True,
            title=device.name().value,
        )
        self.sensor_width = sensor_width
        self.sensor_height = sensor_height
        self.device = device
        self.program = vispy.gloo.Program(VERTEX_SHADER, FRAGMENT_SHADER)
        self.ts_and_ons = np.zeros(
            (self.sensor_width, self.sensor_height),
            dtype=np.float32,
        )
        self.current_t = 0
        self.offset_t = 0
        self.texture = vispy.gloo.texture.Texture2D(
            data=self.ts_and_ons,
            format="luminance",
            internalformat="r32f",
        )
        self.projection = vispy.util.transforms.ortho(
            0, sensor_width, 0, sensor_height, -1, 1
        )
        self.program["u_projection"] = self.projection
        self.program["u_texture"] = self.texture
        self.program["u_t"] = 0
        self.program["u_tau"] = 200000
        self.coordinates = np.zeros(
            4,
            dtype=[("a_position", np.float32, 2), ("a_texcoord", np.float32, 2)],
        )
        self.coordinates["a_position"] = np.array(
            [
                [0, 0],
                [sensor_width, 0],
                [0, sensor_height],
                [sensor_width, sensor_height],
            ]
        )
        self.coordinates["a_texcoord"] = np.array([[0, 0], [0, 1], [1, 0], [1, 1]])
        self.program.bind(vispy.gloo.VertexBuffer(self.coordinates))
        self.projection = vispy.util.transforms.ortho(
            0,
            self.sensor_width,
            0,
            self.sensor_height,
            -1,
            1,
        )
        vispy.gloo.set_clear_color("#292929")
        self.timer = vispy.app.Timer(FRAME_DURATION, connect=self.update, start=True)
        self.show()

    def on_resize(self, event):
        if self.program is not None:
            width, height = event.physical_size
            vispy.gloo.set_viewport(0, 0, width, height)
            self.projection = vispy.util.transforms.ortho(
                0, width, 0, height, -100, 100
            )
            self.program["u_projection"] = self.projection
            ratio = width / float(height)
            sensor_ratio = self.sensor_width / float(self.sensor_height)
            if ratio < sensor_ratio:
                w, h = width, width / sensor_ratio
                x, y = 0, int((height - h) / 2)
            else:
                w, h = height * sensor_ratio, height
                x, y = int((width - w) / 2), 0
            self.coordinates["a_position"] = np.array(
                [[x, y], [x + w, y], [x, y + h], [x + w, y + h]]
            )
            self.program.bind(vispy.gloo.VertexBuffer(self.coordinates))

    def on_draw(self, event):
        assert self.program is not None
        vispy.gloo.clear(color=True, depth=True)
        status, packet = next(device)
        if packet is not None:
            if "dvs_events" in packet:
                assert status.ring is not None and status.ring.current_t is not None
                self.program["u_t"] = np.float32(status.ring.current_t)
                self.ts_and_ons[
                    packet["dvs_events"]["x"], packet["dvs_events"]["y"]
                ] = packet["dvs_events"]["t"].astype(np.float32) * (
                    packet["dvs_events"]["on"].astype(np.float32) * 2.0 - 1.0
                )
            elif status.ring is not None and status.ring.current_t is not None:
                self.program["u_t"] = np.float32(status.ring.current_t)
            if status.ring is not None and status.ring.backlog() > 1000:
                device.clear_backlog(until=0)
        self.texture.set_data(self.ts_and_ons)
        self.program.draw("triangle_strip")


if __name__ == "__main__":
    nd.print_device_list()
    camera_thread: typing.Optional[threading.Thread] = None
    with nd.open(iterator_timeout=FRAME_DURATION) as device:
        print(device.serial(), device.properties())
        canvas = Canvas(
            sensor_width=int(device.properties().width),
            sensor_height=int(device.properties().height),
            device=device,
        )
        vispy.app.run()



import dataclasses
import typing

import numpy


class SerializationError(ValueError):
    """Error raised during Serialization"""

    pass


class DeserializationError(ValueError):
    """Error raised during Deserialization"""

    pass


@dataclasses.dataclass(init=False)
class uint128:
    high: numpy.uint64
    low: numpy.uint64

    def __init__(self, num):
        self.high = numpy.uint64(num >> 64)
        self.low = numpy.uint64(num & 0xFFFFFFFFFFFFFFFF)

    def __int__(self):
        return (int(self.high) << 64) | int(self.low)


@dataclasses.dataclass(init=False)
class int128:
    high: numpy.int64
    low: numpy.uint64

    def __init__(self, num):
        self.high = numpy.int64(num >> 64)
        self.low = numpy.uint64(num & 0xFFFFFFFFFFFFFFFF)

    def __int__(self):
        return (int(self.high) << 64) | int(self.low)


@dataclasses.dataclass(init=False)
class char:
    value: str

    def __init__(self, s):
        if len(s) != 1:
            raise ValueError("`char` expects a single unicode character")
        self.value = s

    def __str__(self):
        return self.value


unit = type(None)
bool = bool
int8 = typing.Union[numpy.int8, int]
int16 = typing.Union[numpy.int16, int]
int32 = typing.Union[numpy.int32, int]
int64 = typing.Union[numpy.int64, int]
uint8 = typing.Union[numpy.uint8, int]
uint16 = typing.Union[numpy.uint16, int]
uint32 = typing.Union[numpy.uint32, int]
uint64 = typing.Union[numpy.uint64, int]
float32 = typing.Union[numpy.float32, int]
float64 = typing.Union[numpy.float64, int]



import datetime
import pathlib

import neuromorphic_drivers as nd

with nd.open(raw=True) as device:
    print(device.properties())
    filename = (
        pathlib.Path(__file__).resolve().parent
        / f"{datetime.datetime.now(tz=datetime.timezone.utc).isoformat().replace('+00:00', 'Z').replace(':', '-')}.raw"
    )
    print(f"writing to {filename}")
    total = 0
    with open(filename, "wb") as output:
        for status, packet in device:
            output.write(packet)
            output.flush()
            total += len(packet)
            print(
                f"{total / 1e6:.1f} MB (backlog: {status.ring.backlog()}, raw: {status.ring.raw_packets()})"
            )



from . import status
from .generated import enums, unions
from .neuromorphic_drivers import Device as ExtensionDevice


class Device(ExtensionDevice):
    def __new__(cls, *args, **kwargs):
        return super().__new__(cls, *args, **kwargs)

    def name(self):
        return enums.Name(super().name())

    def speed(self):
        return enums.Speed(super().speed())

    def properties(self):
        return unions.name_to_properties(self.name())

    def __iter__(self):
        return self

    def __next__(self):
        system_time, ring_status, packet = super().__next__()
        return (
            status.Status(
                system_time=system_time,
                ring=None if ring_status is None else status.RingStatus(*ring_status),
            ),
            packet,
        )

    def update_configuration(self, configuration: unions.Configuration) -> None:
        return ExtensionDevice.update_configuration(
            self, (configuration.type(), configuration.serialize())
        )



import datetime
import pathlib
import time

import neuromorphic_drivers as nd
import numpy as np

nd.print_device_list()

PRIMARY_SERIAL = "00050421"
SECONDARY_SERIAL = "00050423"

secondary_device = nd.open(
    serial=SECONDARY_SERIAL,
    raw=True,
    iterator_timeout=0.1,
    configuration=nd.prophesee_evk4.Configuration(
        nd.prophesee_evk4.Biases(
            diff_on=120,
            diff_off=120,
        ),
        clock=nd.prophesee_evk4.Clock.EXTERNAL,
    ),
)
time.sleep(1.0)  # wait to make sure that secondary is ready
primary_device = nd.open(
    serial=PRIMARY_SERIAL,
    raw=True,
    iterator_timeout=0.1,
    configuration=nd.prophesee_evk4.Configuration(
        nd.prophesee_evk4.Biases(
            diff_on=120,
            diff_off=120,
        ),
        clock=nd.prophesee_evk4.Clock.INTERNAL_WITH_OUTPUT_ENABLED,
    ),
)
devices = [primary_device, secondary_device]

dirname = pathlib.Path(__file__).resolve().parent
name = (
    datetime.datetime.now(tz=datetime.timezone.utc)
    .isoformat()
    .replace("+00:00", "Z")
    .replace(":", "-")
)
outputs = [
    open(dirname / f"{name}_{PRIMARY_SERIAL}.raw", "wb"),
    open(dirname / f"{name}_{SECONDARY_SERIAL}.raw", "wb"),
]

backlogs = np.array([0 for _ in devices])
while True:
    index = np.argmax(backlogs)
    status, packet = devices[index].__next__()
    if status.ring is None:
        backlogs[:] += 1
        backlogs[index] = 0
    else:
        backlog = status.ring.backlog()
        backlogs[:] += 1
        backlogs[index] = backlog
        delay = status.delay()
        if delay is not None:
            if packet is not None:
                print(
                    f"{index}: {round(delay * 1e6)} µs, backlog: {backlog}, bytes: {len(packet)}"
                )
                outputs[index].write(packet)
                outputs[index].flush()
            else:
                print(f"{index}: {round(delay * 1e6)} µs, backlog: {backlog}")



import time

import neuromorphic_drivers as nd

configuration = nd.prophesee_evk4.Configuration(
    biases=nd.prophesee_evk4.Biases(
        diff_off=130,
        diff_on=150,
    )
)

with nd.open(configuration=configuration) as device:
    start = time.monotonic()
    previous_output = start
    count = 0
    updated = False
    for status, packet in device:
        now = time.monotonic()
        if now - previous_output >= 1.0:
            print(f"{count / (now - previous_output):.2f} events/s")
            previous_output = now
            count = 0
        if "dvs_events" in packet:
            count += len(packet["dvs_events"])
        if not updated and now - start > 5.0:
            print("update biases")
            configuration.biases.diff_on -= 50
            configuration.biases.diff_off -= 50
            device.update_configuration(configuration)
            updated = True



import enum


class Speed(enum.Enum):
    UNKNOWN = "USB Unknown speed"
    LOW = "USB 1.0 Low Speed (1.5 Mb/s)"
    FULL = "USB 1.1 Full Speed (12 Mb/s)"
    HIGH = "USB 2.0 High Speed (480 Mb/s)"
    SUPER = "USB 3.0 SuperSpeed (5.0 Gb/s)"
    SUPER_PLUS = "USB 3.1 SuperSpeed+ (10.0 Gb/s)"


class Name(enum.Enum):
    PROPHESEE_EVK3_HD = "Prophesee EVK3 HD"
    PROPHESEE_EVK4 = "Prophesee EVK4"



import pathlib
import struct
import typing
import zlib

import numpy

from . import serde


class Mask:
    def __init__(self, width: serde.type.uint16, height: serde.type.uint16, fill: bool):
        self.x_unpacked = numpy.full(width, fill_value=fill, dtype=bool)
        self.y_unpacked = numpy.full(height, fill_value=fill, dtype=bool)

    def clear_x_range(
        self,
        start: serde.type.uint16,
        stop: serde.type.uint16,
        step: serde.type.uint16,
    ):
        for index in range(start, stop, step):
            self.x_unpacked[index] = False

    def fill_x_range(
        self,
        start: serde.type.uint16,
        stop: serde.type.uint16,
        step: serde.type.uint16,
    ):
        for index in range(start, stop, step):
            self.x_unpacked[index] = True

    def clear_y_range(
        self,
        start: serde.type.uint16,
        stop: serde.type.uint16,
        step: serde.type.uint16,
    ):
        for index in range(start, stop, step):
            self.y_unpacked[index] = False

    def fill_y_range(
        self,
        start: serde.type.uint16,
        stop: serde.type.uint16,
        step: serde.type.uint16,
    ):
        for index in range(start, stop, step):
            self.x_unpacked[index] = True

    def clear_rectangle(
        self,
        x: serde.type.uint16,
        y: serde.type.uint16,
        width: serde.type.uint16,
        height: serde.type.uint16,
    ):
        self.clear_x_range(start=x, stop=x + width, step=1)
        self.clear_y_range(start=y, stop=y + height, step=1)

    def x(self) -> tuple[serde.type.uint64, ...]:
        uint8 = numpy.packbits(self.x_unpacked, bitorder="little")
        if len(uint8) % 8 != 0:
            uint8 = numpy.append(
                uint8, numpy.zeros(8 - len(uint8) % 8, dtype=numpy.uint8)
            )
        return tuple(uint8.view(dtype="=u8").tolist())

    def y(self) -> tuple[serde.type.uint64, ...]:
        uint8 = numpy.packbits(self.y_unpacked, bitorder="little")
        if len(uint8) % 8 != 0:
            uint8 = numpy.append(
                uint8, numpy.zeros(8 - len(uint8) % 8, dtype=numpy.uint8)
            )
        return tuple(uint8.view(dtype="=u8").tolist())

    def pixels(self) -> numpy.ndarray[typing.Any, numpy.dtype[numpy.bool_]]:
        result = numpy.full(
            (len(self.y_unpacked), len(self.x_unpacked)), fill_value=False, dtype=bool
        )
        result[:, self.x_unpacked] = True
        result[numpy.flip(self.y_unpacked), :] = True
        return numpy.bitwise_not(result)

    def write_pixels_to_png_file(self, path: typing.Union[str, bytes, pathlib.Path]):
        def pack(tag: bytes, data: bytes):
            content = tag + data
            return b"".join(
                (
                    struct.pack(">I", len(data)),
                    content,
                    struct.pack(">I", 0xFFFFFFFF & zlib.crc32(content)),
                )
            )

        pixels = self.pixels()
        with open(path, "wb") as file:
            file.write(
                b"".join(
                    [
                        b"\x89PNG\r\n\x1a\n",
                        pack(
                            b"IHDR",
                            struct.pack(
                                ">IIBBBBB",
                                pixels.shape[1],
                                pixels.shape[0],
                                1,
                                0,
                                0,
                                0,
                                0,
                            ),
                        ),
                        pack(
                            b"IDAT",
                            zlib.compress(
                                numpy.hstack(
                                    (
                                        numpy.zeros(
                                            (pixels.shape[0], 1), dtype=numpy.uint8
                                        ),
                                        numpy.packbits(pixels).reshape(
                                            (pixels.shape[0], pixels.shape[1] // 8)
                                        ),
                                    )
                                ).tobytes(),
                                9,
                            ),
                        ),
                        pack(b"IEND", b""),
                    ]
                )
            )



def install_udev_rules():
    import argparse
    import builtins
    import os
    import subprocess
    import sys

    parser = argparse.ArgumentParser(
        prog="neuromorphic-drivers-install-udev-rules",
        description="Install udev rules for Neuromorphic devices",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        "--output",
        "-o",
        default="/etc/udev/rules.d/65-neuromorphic-drivers.rules",
        help="Rules file path",
    )
    parser.add_argument(
        "--skip-system-check",
        "-s",
        action="store_true",
        help="Do not check the operating system type",
    )
    parser.add_argument(
        "--skip-root-check",
        "-r",
        action="store_true",
        help="Do not check for root privileges",
    )
    args = parser.parse_args()
    if not args.skip_system_check:
        if sys.platform == "win32":
            sys.stderr.write("udev rules are not compatible with Windows\n")
            sys.exit(1)
        if sys.platform == "darwin":
            sys.stderr.write("udev rules are not compatible with macOS\n")
            sys.exit(1)
    if not args.skip_root_check:
        if os.geteuid() != 0:
            sys.stderr.write(
                "this program requires root privileges (try running it with sudo)\n"
            )
            sys.exit(1)
    with builtins.open(args.output, "w") as output:
        output.write(
            "".join(
                (
                    'SUBSYSTEM=="usb", ATTRS{idVendor}=="152a",ATTRS{idProduct}=="84[0-1]?", MODE="0666"\n',
                    'SUBSYSTEM=="usb", ATTRS{idVendor}=="04b4",ATTRS{idProduct}=="00f[4-5]", MODE="0666"\n',
                )
            )
        )
    subprocess.check_call("udevadm control --reload-rules")
    subprocess.check_call("udevadm trigger")
